# -*- encoding: utf-8 -*-

# django-fossil: immutable audit trails for interrelated model instances
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from django.db.models.base import ModelBase
from django.db.models import ForeignKey, Field
from django.db.models.fields.related import ReverseSingleRelatedObjectDescriptor
from django.db.models.fields.related import RelatedField
from django.db.models.fields.related import ForeignRelatedObjectsDescriptor
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core import serializers
from django.forms import ModelChoiceField, Select
from django.utils import simplejson

from fossil.models import Fossil

# ------- MODELS -------

class FossilForeignKey(ForeignKey):
    def __init__(self, to, **kwargs):
        """Here we must to store indirect object type to use it later"""

        # Stores indirect model classe relationship
        self.indirect_to = to
        self.indirect_kwargs = {
            'db_index': kwargs.pop('db_index', None),
            }

        super(FossilForeignKey, self).__init__(Fossil, **kwargs)

    def contribute_to_class(self, cls, name):
        super(FossilForeignKey, self).contribute_to_class(cls, name)

        # Replaces field descriptor
        setattr(cls, self.name, FossilSingleObjectDescriptor(self))

    @property
    def indirect_content_type(self):
        if not getattr(self, '_indirect_content_type', None):
            self._indirect_content_type = ContentType.objects.get_for_model(self.indirect_to)
        return self._indirect_content_type

    def pre_save(self, model_instance, add):
        try:
            obj = model_instance._indirect_references.get(self, None)
        except AttributeError:
            obj = None

        if not obj: return None

        # Gets or creates a fossil for current version of the related object
        fossil = Fossil.objects.create_for_object(obj)

        # Returns fossil's id
        return fossil.pk

    def formfield(self, **kwargs):
        defaults = {
                'form_class': FossilChoiceField,
                'indirect_to': self.indirect_to,
                }
        defaults.update(kwargs)

        return super(FossilForeignKey, self).formfield(**defaults)

    def validate(self, value, model_instance):
        # IMPORTANT: This must be to ForeignKey's superclass to ignore its own override
        super(ForeignKey, self).validate(value, model_instance)

        if value is None:
            return

        # This is really seemed to ForeignKey validation, but makes from indirect model class (not fossil revisions)
        qs = self.indirect_to._default_manager.filter(**{self.rel.field_name:value})
        qs = qs.complex_filter(self.rel.limit_choices_to)
        if not qs.exists():
            raise ValidationError(self.error_messages['invalid'] % {
                'model': self.rel.to._meta.verbose_name, 'pk': value})

    def contribute_to_related_class(self, cls, related):
        """This method is based on ForeignKey's method with same name, but have to be
        here to override it and implement the relationship direct to related class,
        instead of Fossil."""

        cls = self.indirect_to

        # Internal FK's - i.e., those with a related name ending with '+' -
        # don't get a related descriptor.
        if not self.rel.is_hidden():
            setattr(cls, related.get_accessor_name(), FossilRelatedObjectDescriptor(related))
        if self.rel.field_name is None:
            self.rel.field_name = cls._meta.pk.name

class DictKeyAttribute(dict):
    """
    Dictionary that allows you access dict key values as attributes
    """
    def __getattr__(self, name):
        try:
            value = self[name]
        except KeyError:
            raise AttributeError

        if isinstance(value, dict):
            value = DictKeyAttribute(value)

        return value

class FossilProxy(object):
    fossil = None
    _object_fossil = None

    def __init__(self, fossil=None):
        if isinstance(fossil, (Fossil, dict)):
            self.fossil = fossil
        elif isinstance(fossil, basestring):
            self.fossil = simplejson.loads(fossil)
        else:
            raise Exception('Invalid fossil source (should be a Fossil instance, JSON string or dictionary)')

    @property
    def object_fossil(self):
        if self._object_fossil is None:
            if isinstance(self.fossil, Fossil):
                self._object_fossil = self.fossil.get_object_fossil()
            else:
                self._object_fossil = DictKeyAttribute(self.fossil)

        return self._object_fossil

    def __getattr__(self, name):
        if name in ('object_fossil','_object_fossil','fossil'): #,'get_attr_from_dict'):
            raise AttributeError, name

        try:
            if isinstance(self.fossil, dict):
                value = self.fossil[name]

                if isinstance(value, dict):
                    value = DictKeyAttribute(value)

                return value
            else:
                return getattr(self.object_fossil.object, name)
        except (AttributeError, KeyError):
            raise AttributeError, name

    def __repr__(self):
        class_name = self.__class__.__name__

        if isinstance(self.fossil, Fossil):
            model_name = self.fossil.content_type.model
            object_id = self.object_fossil.object.pk
        else:
            model_name = self.fossil.get('__model__', 'UnknownModel')
            object_id = self.fossil.get('pk', 'UnknownID')

        return '<%s: %s #%s>'%(class_name, model_name, object_id)

    def __unicode__(self):
        if isinstance(self.fossil, Fossil):
            return unicode(self.fossil)
        else:
            return self.fossil['__unicode__']

    def __str__(self):
        if isinstance(self.fossil, Fossil):
            return str(self.fossil)
        else:
            return self.fossil['__unicode__']

class FossilSingleObjectDescriptor(ReverseSingleRelatedObjectDescriptor):
    """Take more details from the superclass."""

    def __init__(self, field_with_rel):
        self.field = field_with_rel    

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self

        cache_name = self.field.get_cache_name()
        
        try:
            value = getattr(instance, cache_name)
        except AttributeError:
            value = Fossil.objects.get(
                    pk=getattr(instance, self.field.attname)
                    )

        if isinstance(value, Fossil):
            value = FossilProxy(value)
        
        return value
    
    def __set__(self, instance, value):
        """Here we have to store the referenced object in a temporary attribute of
        instance and set a temporary Fossil instance to the field."""

        instance._indirect_references = getattr(instance, '_indirect_references', {})
        instance._indirect_references[self.field] = value

        if value:
            setattr(instance, self.field.attname, value.pk)
            setattr(instance, self.field.get_cache_name(), value)
        else:
            setattr(instance, self.field.attname, None)
            setattr(instance, self.field.get_cache_name(), None)

class FossilRelatedObjectDescriptor(ForeignRelatedObjectsDescriptor):
    def create_manager(self, instance, superclass):
        """This function is not so full as Django does. If we have necessity we can enhance
        it to support creating and clearing objects."""

        rel_field = self.related.field
        rel_model = self.related.model

        conditions = {
                '%s__content_type' % rel_field.name: rel_field.indirect_content_type,
                '%s__object_id' % rel_field.name: instance.pk,
                }

        queryset = rel_model._default_manager.filter(**conditions)

        return queryset

# ------- FORMS -------

class FossilChoiceWidget(Select):
    """Form Widget class mostly used by FossilForeignKey fields."""

    def render(self, name, value, *args, **kwargs):
        # Swap initial value to pk of real reference object
        if value:
            fossil = Fossil.objects.filter(pk=value).values_list('object_id', flat=True)[0]
            value = fossil[0]

        return super(FossilChoiceWidget, self).render(name, value, *args, **kwargs)

class FossilChoiceField(ModelChoiceField):
    """Form Field class mostly used by FossilForeignKey fields."""

    widget = FossilChoiceWidget
    indirect_to = None

    def __init__(self, queryset, *args, **kwargs):
        self.indirect_to = kwargs.pop('indirect_to', None)

        # Swap Fossil queryset for foreign model class queryset
        if self.indirect_to:
            queryset = self.indirect_to.objects.all()
            if self.indirect_to._meta.ordering:
                queryset = queryset.order_by(*self.indirect_to._meta.ordering)
            else:
                queryset = queryset.order_by('pk')
        else:
            queryset = None

        super(FossilChoiceField, self).__init__(queryset, *args, **kwargs)





# -*- encoding: utf-8 -*-

# django-fossil: immutable audit trails for interrelated model instances
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import hashlib, datetime

from django.db import models
from django.core.serializers import serialize, deserialize
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.generic import GenericForeignKey

try:
    set
except:
    from sets import Set as set

class FossilManager(models.Manager):
    def create_for_object(self, obj):
        if hasattr(obj, 'serialize_for_fossil'):
            serialized = obj.serialize_for_fossil()
        else:
            serialized = serialize('json', [obj])

        hash_key = hashlib.sha256(serialized).hexdigest()

        return Fossil.objects.get_or_create(
                pk=hash_key,
                defaults={
                    'serialized': serialized,
                    'display_text': unicode(obj),
                    'content_type': ContentType.objects.get_for_model(obj),
                    'object_id': unicode(obj.pk),
                    },
                )[0]

    def fossils_of_object(self, obj):
        """
        Returns a list of fossils of a given object, ordered by date
        """
        qs = self.get_query_set()
        c_type = ContentType.objects.get_for_model(obj)

        return qs.filter(content_type=c_type, object_id=obj.pk).order_by('creation')

    def indexed(self, **kwargs):
        """
        Find fossils by fossil indexes
        """
        qs = self.get_query_set()

        # Find all indexes by given key and value
        indexers_qs = FossilIndexer.objects.all()
        indexers = []
        for k,v in kwargs.items():
            # Field is the first node before '__' if there is one in the key (k)
            if '__' in k:
                field, lookup = k.split('__',1)
                filters = {'key': field, 'value__'+lookup: v}
            else:
                filters = {'key': k, 'value': v}

            # Appends lists of found fossil indexers (only their fossil PKs)
            indexers.append(set(indexers_qs.filter(**filters).distinct().values_list('fossil', flat=True)))

        # This reduce makes an intersection of all indexers, this means that only PKs found in
        # ALL indexers will be in new list 'pks'
        pks = reduce(lambda a,b: a.intersection(b), indexers)

        return qs.filter(pk__in=pks)

class Fossil(models.Model):
    objects = _default_manager = FossilManager()

    id = models.CharField(max_length=64, primary_key=True) # Hash key
    serialized = models.TextField(blank=True, default='')
    display_text = models.TextField(blank=True, default='')
    creation = models.DateTimeField(blank=True, default=datetime.datetime.now)
    content_type = models.ForeignKey(ContentType)
    object_id = models.TextField()
    object = GenericForeignKey()
    is_most_recent = models.BooleanField(blank=True, default=True, db_index=True)
    previous_revision = models.ForeignKey('self', null=True, blank=True)
    revision_sequential = models.PositiveIntegerField(null=True, blank=True, db_index=True)

    def __unicode__(self):
        return self.display_text

    def get_object_fossil(self):
        """
        Returns the stored version of this object.
        """

        data = self.serialized

        if isinstance(data, unicode):
            data = data.encode("utf8")

        manager = self.content_type.model_class().objects
        if hasattr(manager, 'deserialize_for_fossil'):
            return manager.deserialize_for_fossil(data)
        else:
            return list(deserialize('json', data))[0].object

    def set_indexer(self, key, value):
        """
        Creates (or updates if exists) a fossil index for this fossil + given key and value
        """
        indexer, new = FossilIndexer.objects.get_or_create(
                fossil=self,
                key=key,
                defaults=dict(value=value),
                )

        if not new and unicode(value) != indexer.value:
            indexer.value = value
            indexer.save()

    def unset_indexer(self, key):
        """Removes an indexer by a given key"""
        self.indexers.filter(key=key).delete()
    
class FossilIndexerManager(models.Manager):
    def key(self, key, fail_silent=False):
        try:
            qs = self.get_query_set()
            return qs.get(key=key)
        except FossilIndexer.DoesNotExist:
            if fail_silent:
                return None
            else:
                raise

class FossilIndexer(models.Model):
    """
    Class used to index fossil by field values. This is a sollution for querying
    fossils withouth use search in the field 'serialized'. Of course, this is
    because index + join is faster than like.
    """
    class Meta:
        unique_together = (
                ('fossil','key','value'),
                )

    objects = _default_manager = FossilIndexerManager()

    fossil = models.ForeignKey('Fossil', related_name='indexers')
    key = models.CharField(max_length=250)
    value = models.CharField(max_length=250, null=True)

    def __unicode__(self):
        return u'%s = %s'%(self.key, self.value)

# SIGNALS
from django.db.models import signals

def fossil_post_save(sender, instance, signal, **kwargs):
    # Updates old revisions setting them with "is_most_recent" as False
    if instance.is_most_recent:
        Fossil.objects.filter(
                content_type=instance.content_type,
                object_id=instance.object_id,
                creation__lt=instance.creation,
                is_most_recent=True,
                ).exclude(
                        pk=instance.pk,
                        ).update(
                                is_most_recent=False,
                                )

    # Gets "previous_revision" from last one
    if not instance.previous_revision:
        try:
            instance.previous_revision = Fossil.objects.filter(
                content_type=instance.content_type,
                object_id=instance.object_id,
                creation__lt=instance.creation,
                ).exclude(
                    pk=instance.pk,
                    ).latest('creation')
            instance.save()
        except Fossil.DoesNotExist:
            pass

    # Gets next "revision_sequential"
    if not instance.revision_sequential:
        if not instance.previous_revision:
            instance.revision_sequential = 1
            instance.save()
        elif instance.previous_revision.revision_sequential:
            instance.revision_sequential = instance.previous_revision.revision_sequential + 1
            instance.save()

signals.post_save.connect(fossil_post_save, sender=Fossil)




import datetime

from django.db import models
from django.contrib.auth.models import User

from fossil.fields import FossilForeignKey
from fossil.models import Fossil

class Supplier(models.Model):
    name = models.CharField(max_length=100)
    location = models.TextField(blank=True)
    salary = models.DecimalField(max_digits=12, decimal_places=2, null=True, blank=True)
    starred = models.BooleanField(default=False, blank=True)
    points = models.IntegerField(default=0, blank=True)
    date_foundation = models.DateField(blank=True, null=True)

    def __unicode__(self):
        return self.name

    def fossil_revisions(self):
        return Fossil.objects.fossils_of_object(self)

class Purchase(models.Model):
    date = models.DateTimeField(blank=True, default=datetime.datetime.now)
    supplier = FossilForeignKey(Supplier, null=True, blank=True)
    user = models.ForeignKey(User, null=True, blank=True)

    def fossil_revisions(self):
        return Fossil.objects.fossils_of_object(self)

from django.db.models import signals

def purchase_post_save(instance, **kwargs):
    fossil = Fossil.objects.create_for_object(instance)

    fossil.set_indexer('pk', instance.pk)
    fossil.set_indexer('date', instance.date.strftime('%d/%m/%Y'))

signals.post_save.connect(purchase_post_save, sender=Purchase)




from django.contrib import admin

from models import Supplier, Purchase

class AdminSupplier(admin.ModelAdmin):
    pass

class AdminPurchase(admin.ModelAdmin):
    list_display = ('date','supplier','user')

admin.site.register(Supplier, AdminSupplier)
admin.site.register(Purchase, AdminPurchase)




import datetime

from django.db import models, connection
from django.core.management.color import no_style
from django.contrib.flatpages.models import FlatPage
from django.contrib.contenttypes import generic
from django.db.models.sql.query import setup_join_cache
from django.contrib.auth.models import User

from fossil.fields import FossilForeignKey

class Supplier(models.Model):
    class Meta:
        app_label = 'reversion_relations'

    name = models.CharField(max_length=100)
    location = models.TextField(blank=True)
    salary = models.DecimalField(max_digits=12, decimal_places=2, null=True, blank=True)
    starred = models.BooleanField(default=False, blank=True)
    points = models.IntegerField(default=0, blank=True)
    date_foundation = models.DateField(blank=True, null=True)

    def __unicode__(self):
        return self.name

class Purchase(models.Model):
    class Meta:
        app_label = 'reversion_relations'

    date = models.DateTimeField(blank=True, default=datetime.datetime.now)
    supplier = FossilForeignKey(Supplier, null=True, blank=True)
    user = models.ForeignKey(User, null=True, blank=True)

setup_join_cache(Supplier)
setup_join_cache(Purchase)

def create_tables():
    cursor = connection.cursor()
    style = no_style()
    tables = connection.introspection.table_names()
    seen_models = connection.introspection.installed_models(tables)

    sql, references = connection.creation.sql_create_model(Supplier, style, seen_models)
    new_sql, new_ref = connection.creation.sql_create_model(Purchase, style, seen_models)
    sql.extend(new_sql); references.update(new_ref)

    pending_references = {}
    for refto, refs in references.items():
        pending_references.setdefault(refto, []).extend(refs)
        if refto in seen_models:
            sql.extend(connection.creation.sql_for_pending_references(refto, style, pending_references))
    sql.extend(connection.creation.sql_for_pending_references(Supplier, style, pending_references))
    sql.extend(connection.creation.sql_for_pending_references(Purchase, style, pending_references))
    for statement in sql:
        cursor.execute(statement)




# -*- encoding: utf-8 -*-

# django-fossil: immutable audit trails for interrelated model instances
#
# Copyright (C) 2010 BIREME/PAHO/WHO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

VERSION = (0, 7, 'stable')

def get_version():
    return '%d.%d-%s'%VERSION

__author__ = 'OpenTrials team (Marinho Brandao, Luciano Ramalho and Rafael Soares)'
#__date__ = '$Date: 2008-07-26 14:04:51 -0300 (Ter, 26 Fev 2008) $'[7:-2]
__license__ = 'GNU Lesser General Public License (LGPL)'
__url__ = 'http://reddes.bvsalud.org/projects/clinical-trials/browser/sandbox/django-fossil'
__version__ = get_version()




# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding field 'Fossil.revision_sequential'
        db.add_column('fossil_fossil', 'revision_sequential', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True, null=True, blank=True), keep_default=False)


    def backwards(self, orm):
        
        # Deleting field 'Fossil.revision_sequential'
        db.delete_column('fossil_fossil', 'revision_sequential')


    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'fossil.fossil': {
            'Meta': {'object_name': 'Fossil'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'creation': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now', 'blank': 'True'}),
            'display_text': ('django.db.models.fields.TextField', [], {'default': "''", 'blank': 'True'}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '64', 'primary_key': 'True'}),
            'is_most_recent': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'object_id': ('django.db.models.fields.TextField', [], {}),
            'previous_revision': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['fossil.Fossil']", 'null': 'True', 'blank': 'True'}),
            'revision_sequential': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'serialized': ('django.db.models.fields.TextField', [], {'default': "''", 'blank': 'True'})
        },
        'fossil.fossilindexer': {
            'Meta': {'unique_together': "(('fossil', 'key', 'value'),)", 'object_name': 'FossilIndexer'},
            'fossil': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'indexeds'", 'to': "orm['fossil.Fossil']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '250'}),
            'value': ('django.db.models.fields.CharField', [], {'max_length': '250'})
        }
    }

    complete_apps = ['fossil']



import os, sys

# Downloads setuptools if not find it before try to import
import ez_setup
ez_setup.use_setuptools()

from setuptools import setup
import fossil

setup(
    name = 'django-fossil',
    version = fossil.__version__,
    description = "Fossil is a library to store fossilized tracks of objects from Django's ORM",
    author = fossil.__author__,
    author_email = 'opentrials-dev@listas.bireme.br',
    url = fossil.__url__,
    license = fossil.__license__,
    packages = ['fossil'],
)




# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'Fossil'
        db.create_table('fossil_fossil', (
            ('id', self.gf('django.db.models.fields.CharField')(max_length=64, primary_key=True)),
            ('serialized', self.gf('django.db.models.fields.TextField')(default='', blank=True)),
            ('display_text', self.gf('django.db.models.fields.TextField')(default='', blank=True)),
            ('creation', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now, blank=True)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal('fossil', ['Fossil'])


    def backwards(self, orm):
        
        # Deleting model 'Fossil'
        db.delete_table('fossil_fossil')


    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'fossil.fossil': {
            'Meta': {'object_name': 'Fossil'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'creation': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now', 'blank': 'True'}),
            'display_text': ('django.db.models.fields.TextField', [], {'default': "''", 'blank': 'True'}),
            'id': ('django.db.models.fields.CharField', [], {'max_length': '64', 'primary_key': 'True'}),
            'object_id': ('django.db.models.fields.TextField', [], {}),
            'serialized': ('django.db.models.fields.TextField', [], {'default': "''", 'blank': 'True'})
        }
    }

    complete_apps = ['fossil']
